"""
Tests ROS2 and stdout using the talker/listened nodes
"""

import pytest

pytest.importorskip("rclpy")

import asyncio
import logging
import subprocess
import sys
from os import environ
from typing import Any, AsyncGenerator, Callable, Generator

import rclpy
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_cancellation,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)
from rclpy.qos import QoSProfile
from std_msgs.msg import String

import asyncio_for_robotics.ros2 as afor
import asyncio_for_robotics.textio as afor_io
from asyncio_for_robotics.core._logger import setup_logger

is_win = (
    sys.version_info[0] == 3
    and sys.version_info[1] >= 8
    and sys.platform.startswith("win")
)

pytestmark = pytest.mark.skipif(is_win, reason="Requires a special EvenLoop")

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module")
def zenoh_router():
    """ """
    if environ.get("RMW_IMPLEMENTATION") != "rmw_zenoh_cpp":
        yield None
        return
    logger.info("Starting zenoh router")
    proc = subprocess.Popen(
        ["ros2", "run", "rmw_zenoh_cpp", "rmw_zenohd"],
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        text=True,
    )
    for line in iter(proc.stdout.readline, ""):
        if "Started Zenoh router with id" in line:
            logger.info("Router started!")
            break
    yield proc
    logger.info("Closing zenoh router")
    proc.terminate()
    proc.wait()
    logger.info("Closed zenoh router")


@pytest.fixture(scope="module", autouse=True)
def session(zenoh_router) -> Generator[afor.BaseSession, Any, Any]:
    logger.info("Starting rclpy and session")
    rclpy.init()
    ses = afor.auto_session()
    yield ses
    logger.info("closing rclpy and session")
    ses.close()
    rclpy.shutdown()


@pytest.fixture(scope="module")
def node_process(zenoh_router):
    """Fixture running the ros example publisher.

    This will start a background OS process, then close it on cleanup.
    """
    logger.info("Starting ros node")
    proc = subprocess.Popen(  # runs an OS process in the background
        ["ros2", "run", "demo_nodes_py", "listener"],
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        text=True,
    )
    yield proc
    # cleanup: we close the process
    logger.info("Closing ros node")
    proc.terminate()
    proc.wait()
    logger.info("Closed ros node")


TOPIC = afor.TopicInfo(
    topic="/chatter",
    msg_type=String,
    qos=QoSProfile(
        depth=10000,
    ),
)


@pytest.fixture(scope="module")
def pub(session: afor.BaseSession) -> Generator[Callable[[str], None], Any, Any]:
    with session.lock() as node:
        publisher = node.create_publisher(*TOPIC.as_arg())

    def write_in_proc(input: str) -> None:
        publisher.publish(String(data=input))

    yield write_in_proc
    with session.lock() as node:
        node.destroy_publisher(publisher)


@pytest.fixture
async def sub(
    node_process: subprocess.Popen[str],
) -> AsyncGenerator[afor_io.Sub[str], Any]:
    assert node_process.stdout is not None

    def pre_process(input: str):
        if "I heard: " not in input:
            return None
        input = input.strip()
        input = input.split("I heard: [")[1]
        input = input.removesuffix("]")
        return input

    s = afor_io.from_proc_stdout(node_process, pre_process)
    yield s
    s.close()


@pytest.fixture(autouse=True)
async def listener_ready(
    pub: Callable[[str], None],
    sub: afor_io.Sub[str],
):
    """Returns once the listener node is ready."""
    payload_prefix = "ready_"
    latest_payload = payload_prefix + "0"

    async def periodic_publish():
        nonlocal latest_payload
        counter = 0
        while 1:
            counter += 1
            latest_payload = payload_prefix + str(counter)
            pub(latest_payload)
            await asyncio.sleep(0.1)

    async def wait_for_ready():
        nonlocal latest_payload
        async for msg in sub.listen_reliable(fresh=True):
            if latest_payload in msg:
                return

    periodic_task = asyncio.create_task(periodic_publish())
    ready_task = asyncio.create_task(wait_for_ready())

    try:
        await asyncio.wait_for(ready_task, timeout=5)
    except TimeoutError:
        pytest.fail("Listener node process not ready in time.")
    finally:
        ready_task.cancel()
        periodic_task.cancel()
        sub.alive.clear()
        sub._value_flag.clear()
        sub._value = None



import asyncio
import logging
import subprocess
import sys
import threading
import time
from typing import IO, Callable, TypeVar, Union
import warnings

from asyncio_for_robotics.core.sub import BaseSub

_MsgType = TypeVar("_MsgType", str, bytes)
logger = logging.getLogger(__name__)


def default_preprocess(input: _MsgType) -> _MsgType:
    return input.strip()


class Sub(BaseSub[_MsgType]):
    def __init__(
        self,
        stream: IO[_MsgType],
        pre_process: Callable[[_MsgType], Union[_MsgType, None]] = default_preprocess,
    ) -> None:
        """Subscriber streaming updates of a file.

        Notably useful to subscribe to text printed to stdout of a process.

        Args:
            stream: IO - Generic base class for TextIO and BinaryIO.
            pre_process: Function applied before any other processing. If
                returns None, line is skipped.
        """
        self.stream = stream
        self.pre_process = pre_process
        super().__init__()
        if sys.platform.startswith("win"):
            warnings.warn("Windows requires changing the asyncio loop type")
        else:
            self._event_loop.add_reader(self.stream.fileno(), self._io_update_cbk)

    @property
    def name(self) -> str:
        return f"sub io-{self.stream.name}"

    def _win_io_update_cbk(self, line):
        healthy = True
        if line is not None:
            healthy = self.input_data(line)
        if not healthy:
            return

    def _io_update_cbk(self):
        """Is called on updates to the IO file."""
        line = self.pre_process(self.stream.readline())
        healthy = True
        if line is not None:
            healthy = self.input_data(line)
        if not healthy:
            self.close()

    def close(self):
        """Closes the file reader (not the file)."""
        if self._closed.is_set():
            return
        self._event_loop.remove_reader(self.stream.fileno())
        super().close()

def from_proc_stdout(
    process: subprocess.Popen[_MsgType],
    pre_process: Callable[[_MsgType], Union[_MsgType, None]] = default_preprocess,
) -> Sub[_MsgType]:
    """Creates a textio sub, relaying the lines printed in the process stdout.

    Automatically closes the sub when the process finishes

    Args:
        process: process on which to grab stdout
        pre_process: Function applied before any other processing. If
            returns None, line is skipped.

    Returns:
        textio sub of stdout.
    """
    if process.stdout is None:
        raise TypeError(
            "process.stdout is None. Please use `stdout=subprocess.PIPE` when calling Popen."
        )
    stdout: IO[_MsgType] = process.stdout
    sub = Sub(stdout, pre_process)

    async def close_reader():
        # this blocks the loop destruction somehow
        # await asyncio.to_thread(process.wait)
        while process.poll() is None:
            await asyncio.sleep(1)
        logger.debug(f"{sub.name} closed because of process end.")
        sub.close()

    proc_wait_task = asyncio.create_task(close_reader())

    async def closed_so_stop_waiting():
        nonlocal proc_wait_task
        await sub._closed.wait()
        logger.debug(f"{sub.name} closed cancelling process monitoring task")
        proc_wait_task.cancel()
        await proc_wait_task

    closed_wait_task = asyncio.create_task(closed_so_stop_waiting())
    return sub



from .session import (
    set_auto_session,
    auto_session,
)
from .sub import Sub
from .. import soft_timeout, soft_wait_for, Rate, ConverterSub

__all__ = [
    "soft_wait_for",
    "soft_timeout",
    "Rate",
    "auto_session",
    "set_auto_session",
    "Sub",
    "ConverterSub",
]



from .. import ConverterSub, Rate, soft_timeout, soft_wait_for
from .sub import Sub, from_proc_stdout

__all__ = [
    "from_proc_stdout",
    "soft_wait_for",
    "soft_timeout",
    "Rate",
    "ConverterSub",
    "Sub",
    "from_proc_stdout",
]



"""
Example showing how to use asyncio_for_robotics subscriptions with ROS 2
subscription event callbacks (e.g. matched publishers, lost messages).

afor's generic `BaseSub` is used to accumulate a datastream not natively
supported by `afor`. The external ROS object calls `BaseSub.input_data` which
is threadsafe

Run:
    python3 -m asyncio_for_robotics.example.ros2_event_callback

In another terminal:
    ros2 topic pub /example std_msgs/msg/String "data: hey"

Expected output:
    Match Event:
    New publisher
    {'current_count': 1,
     'current_count_change': 1,
     'total_count': 1,
     'total_count_change': 1}

    I heard: hey
    ...
    I heard: hey

    Match Event:
    Undeclared publisher
    {'current_count': 0,
     'current_count_change': -1,
     'total_count': 1,
     'total_count_change': 0}
"""

import asyncio
import inspect
from contextlib import suppress
from pprint import pprint
from typing import Optional

import rclpy
from colorama import Fore
from rclpy.event_handler import (
    IncompatibleTypeInfo,
    QoSLivelinessChangedInfo,
    QoSMessageLostInfo,
    QoSRequestedDeadlineMissedInfo,
    QoSRequestedIncompatibleQoSInfo,
    QoSSubscriptionMatchedInfo,
    SubscriptionEventCallbacks,
)
from rclpy.qos import QoSProfile
from rclpy.subscription import Subscription
from std_msgs.msg import String

from asyncio_for_robotics.core.sub import BaseSub, _MsgType
from asyncio_for_robotics.ros2 import (
    QOS_DEFAULT,
    BaseSession,
    Sub,
    TopicInfo,
    auto_session,
)

TOPIC = TopicInfo(
    "example",
    String,
)


class AdvancedSub(Sub[_MsgType]):
    def __init__(
        self,
        msg_type: type[_MsgType],
        topic: str,
        qos_profile: QoSProfile = QOS_DEFAULT,
        session: Optional[BaseSession] = None,
        event_callbacks: Optional[SubscriptionEventCallbacks] = None,
    ) -> None:
        """
        Extension of `Sub` that allows passing ROS 2
        `SubscriptionEventCallbacks` when creating the underlying rclpy
        subscription.
        """
        self.event_callbacks = event_callbacks
        super().__init__(msg_type, topic, qos_profile, session)

    def _resolve_sub(self, topic_info: TopicInfo) -> Subscription:
        """
        Create the underlying rclpy Subscription with event callbacks attached.
        """
        with self.session.lock() as node:
            return node.create_subscription(
                **topic_info.as_kwarg(),
                callback=self.callback_for_sub,
                event_callbacks=self.event_callbacks,
            )


async def verbose_match_event(match_sub: BaseSub[QoSSubscriptionMatchedInfo]):
    """Listens to match/unmatch events and print their full contents."""
    async for event in match_sub.listen_reliable():
        event: QoSSubscriptionMatchedInfo
        print("")
        print(f"{Fore.RED}Match Event triggered:{Fore.RESET}")
        if event.current_count_change > 0:
            print(f"{Fore.GREEN}Match publisher{Fore.RESET}")
        if event.current_count_change < 0:
            print(f"{Fore.YELLOW}Unmatch publisher{Fore.RESET}")
        pprint(
            {
                name: value
                for name, value in inspect.getmembers(event)
                if not name.startswith("_")
            }
        )
        print("")


async def what_do_you_hear(sub: Sub[String]):
    """Simply print the payloads of a topic."""
    async for msg in sub.listen_reliable():
        print(f"I heard: {msg.data}")


async def event_example():
    """
    - Creates `BaseSub` instances to receive ROS events in afor.
    - Registers them as ROS 2 subscription event callbacks.
    - Runs a simple String subscription.
    """
    # afor's generic `BaseSub` is used to accumulate a generic datastream. The
    # event callback calls `BaseSub.input_data` which is threadsafe
    lost_sub: BaseSub[QoSMessageLostInfo] = BaseSub()
    match_sub: BaseSub[QoSSubscriptionMatchedInfo] = BaseSub()
    event_cbk = SubscriptionEventCallbacks(
        message_lost=lost_sub.input_data,  # type: ignore
        matched=match_sub.input_data,  # type: ignore
        # more event types can be added
    )
    sub = AdvancedSub(
        **TOPIC.as_kwarg(),
        event_callbacks=event_cbk,
    )
    hear_task = asyncio.create_task(what_do_you_hear(sub))
    match_task = asyncio.create_task(verbose_match_event(match_sub))
    print(
        f"""Start/stop publishing using 
        `ros2 topic pub /{sub.topic_info.topic} std_msgs/msg/String "data: hey"`"""
    )
    await asyncio.wait([hear_task, match_task])


if __name__ == "__main__":
    rclpy.init()
    try:
        with suppress(KeyboardInterrupt, asyncio.CancelledError):
            asyncio.run(event_example())
    finally:
        auto_session().close()
        rclpy.shutdown()



import pytest

from asyncio_for_robotics.core.sub import ConverterSub

pytest.importorskip("zenoh")
import asyncio
import logging
from contextlib import suppress
from typing import Any, AsyncGenerator, Callable, Generator, Optional, Union

import zenoh
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_cancellation,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

from asyncio_for_robotics.core import BaseSub
from asyncio_for_robotics.zenoh import (
    Sub,
    auto_session,
    set_auto_session,
    soft_timeout,
    soft_wait_for,
)

logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[zenoh.Session, Any, Any]:
    set_auto_session(zenoh.open(zenoh.Config()))
    ses = auto_session()
    yield ses
    if not auto_session().is_closed():
        with suppress(zenoh.ZError):
            ses.close()


@pytest.fixture
def pub(session) -> Generator[Callable[[str], None], Any, Any]:
    pub_topic = "test/something"
    logger.debug("Creating PUB-%s", pub_topic)
    p: zenoh.Publisher = auto_session().declare_publisher(
        pub_topic, reliability=zenoh.Reliability.RELIABLE
    )

    def pub_func(input: str):
        p.put(input.encode())

    yield pub_func
    if not auto_session().is_closed():
        logger.debug("closing PUB-%s", pub_topic)
        p.undeclare()

@pytest.fixture
async def sub(session) -> AsyncGenerator[BaseSub[str], Any]:
    inner_sub = Sub("test/**")
    s: BaseSub[str] = ConverterSub(inner_sub, lambda sample: sample.payload.to_string())
    yield s
    s.close()



import asyncio
import logging
from typing import Any, AsyncGenerator, Callable, Generator

import pytest
from base_tests import (
    test_freshness,
    test_listen_one_by_one,
    test_listen_too_fast,
    test_reliable_extremely_fast,
    test_reliable_one_by_one,
    test_reliable_too_fast,
    test_wait_cancellation,
    test_wait_for_value,
    test_wait_new,
    test_wait_next,
)

import asyncio_for_robotics.core as afor
from asyncio_for_robotics.core._logger import setup_logger

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture
async def sub() -> AsyncGenerator[afor.sub.BaseSub[str], Any]:
    s = afor.sub.BaseSub()
    yield s
    s.close()


@pytest.fixture
def pub(sub: afor.sub.BaseSub[str]) -> Generator[Callable[[str], None], Any, Any]:
    def write_in_proc(input: str) -> None:
        sub._input_data_asyncio(input)

    yield write_in_proc



import pytest

pytest.importorskip("rclpy")

import asyncio
import copy
import logging
import sys
from typing import Any, AsyncGenerator, Callable, Generator

import rclpy
from rclpy.publisher import Publisher
from rclpy.qos import QoSProfile
from std_msgs.msg import String
from std_srvs.srv import SetBool

import asyncio_for_robotics as afor
import asyncio_for_robotics.ros2 as aros
from asyncio_for_robotics.core._logger import setup_logger
from asyncio_for_robotics.core.sub import BaseSub
from asyncio_for_robotics.ros2.service import Client, Responder, Server
from asyncio_for_robotics.ros2.session import ThreadedSession, set_auto_session

setup_logger(debug_path="tests")
logger = logging.getLogger("asyncio_for_robotics.test")


@pytest.fixture(scope="module", autouse=True)
def session() -> Generator[aros.BaseSession, Any, Any]:
    logger.info("Starting rclpy and session")
    rclpy.init()
    set_auto_session(ThreadedSession())
    ses = aros.auto_session()
    yield ses
    logger.info("closing rclpy and session")
    ses.close()
    rclpy.shutdown()


topic = aros.TopicInfo(
    "test/srv",
    SetBool,
    QoSProfile(
        depth=500,
    ),
)


@pytest.fixture
async def server(
    session: aros.BaseSession,
) -> AsyncGenerator[Server[SetBool.Request, SetBool.Response], Any]:
    server = Server(**topic.as_kwarg())
    yield server
    server.close()


@pytest.fixture
async def client(
    session: aros.BaseSession,
) -> AsyncGenerator[Client[SetBool.Request, SetBool.Response], Any]:
    client = Client(**topic.as_kwarg())
    yield client
    client.close()


async def test_client_wait_fails_properly(
    client: Client[SetBool.Request, SetBool.Response],
):
    res = await afor.soft_wait_for(client.wait_for_service(), 1)
    assert isinstance(res, TimeoutError), f"No server is avaible, should timeout"


async def test_client_wait(
    server: Server[SetBool.Request, SetBool.Response],
    client: Client[SetBool.Request, SetBool.Response],
):
    res = await afor.soft_wait_for(client.wait_for_service(), 1)
    assert not isinstance(res, TimeoutError), f"Server is avaible, should not timeout"


async def test_client_receives_response(
    server: Server[SetBool.Request, SetBool.Response],
    client: Client[SetBool.Request, SetBool.Response],
):
    response_async = client.call(SetBool.Request(data=True))

    responder = await afor.soft_wait_for(server.wait_for_value(), 1)
    assert not isinstance(responder, TimeoutError), f"Server did not receive request"
    assert responder.request.data == True
    responder.response.success = True
    responder.response.message = "hello"
    responder.send()

    response = await afor.soft_wait_for(response_async, 1)
    assert not isinstance(response, TimeoutError), f"Client did not receive reply"
    assert response.message == "hello"
    assert response.success == True
    return


@pytest.fixture
def pub(
    client: Client[SetBool.Request, SetBool.Response],
) -> Generator[Callable[[bool], None], Any, Any]:

    def send_payload(input: bool) -> None:
        req = SetBool.Request(data=input)
        client.call(req)

    yield send_payload


@pytest.fixture
async def sub(
    server: Server[SetBool.Request, SetBool.Response],
) -> AsyncGenerator[BaseSub[bool], Any]:
    s: BaseSub[bool] = BaseSub()

    def transmit(msg: Responder[SetBool.Request, SetBool.Response]):
        nonlocal s
        msg.send()
        s._input_data_asyncio(msg.request.data)

    server.asap_callback.append(transmit)
    yield s
    server.asap_callback.remove(transmit)


def binary(n: int):
    return n % 2 == 0


async def test_wait_for_value(pub: Callable[[bool], None], sub: BaseSub[bool]):
    logger.info("entered test")
    payload = True
    logger.info("publishing")
    pub(payload)
    logger.info("awaiting data")
    sample = await afor.soft_wait_for(sub.wait_for_value(), 1)
    assert not isinstance(sample, TimeoutError), f"Did not receive response in time"
    logger.info("got data")
    logger.info(sample)
    assert sample == payload
    logger.info("passed")


async def test_wait_new(pub: Callable[[bool], None], sub: BaseSub[bool]):
    payload = True
    pub(payload)
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_new()
    new_sample = await afor.soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_new()
    pub(payload)
    new_sample = await afor.soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample == payload


async def test_wait_next(pub: Callable[[bool], None], sub: BaseSub[bool]):
    first_payload = True
    pub(first_payload)
    sample = await sub.wait_for_value()
    assert not isinstance(sample, TimeoutError), f"Should get a message"

    wait_task = sub.wait_for_next()
    new_sample = await afor.soft_wait_for(wait_task, 0.1)
    assert isinstance(new_sample, TimeoutError), f"Should not get a message"

    wait_task = sub.wait_for_next()
    pub(first_payload)
    for other_payload in range(10):
        await asyncio.sleep(0.001)
        pub(other_payload % 2 == 0)

    new_sample = await afor.soft_wait_for(wait_task, 0.1)
    assert not isinstance(new_sample, TimeoutError), f"Should get the message"
    assert new_sample == first_payload


async def test_listen_one_by_one(pub: Callable[[bool], None], sub: BaseSub[bool]):
    last_payload = True
    pub(last_payload)
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen():
        sample_count += 1
        assert sample == last_payload
        if sample_count >= max_iter:
            break
        last_payload = binary(sample_count)
        pub(last_payload)
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_listen_too_fast(pub: Callable[[bool], None], sub: BaseSub[bool]):
    if sys.platform.startswith("win"):
        sleep_time = 0.1
    else:
        sleep_time = 0.001
    last_payload = True
    pub(last_payload)
    sample_count = 0
    put_count = 2
    max_iter = 10
    await asyncio.sleep(0.01)
    async for sample in sub.listen():
        sample_count += 1
        assert sample == last_payload, f"failed on {sample_count=}"
        if sample_count >= max_iter:
            break
        last_payload = binary(sample_count)
        pub(last_payload)
        put_count += 1
        await asyncio.sleep(0.001)
        last_payload = binary(sample_count)
        pub(last_payload)
        put_count += 1
        await asyncio.sleep(sleep_time)

    assert sample_count == max_iter


async def test_reliable_one_by_one(pub: Callable[[bool], None], sub: BaseSub[bool]):
    last_payload = True
    pub(last_payload)
    sample_count = 0
    put_count = 1
    max_iter = 20
    async for sample in sub.listen_reliable():
        sample_count += 1
        assert sample == last_payload
        if sample_count >= max_iter:
            break
        last_payload = binary(sample_count)
        pub(last_payload)
        put_count += 1

    assert put_count == sample_count == max_iter


async def test_reliable_too_fast(pub: Callable[[bool], None], sub: BaseSub[bool]):
    data = [binary(n) for n in range(30)]
    put_queue: list[bool] = data.copy()
    put_queue.reverse()
    received_buf = []
    listener = sub.listen_reliable(fresh=True, queue_size=len(data) * 2)
    await asyncio.sleep(0.001)
    pub(put_queue.pop())
    async with afor.soft_timeout(2):
        async for sample in listener:
            payload = int(sample)
            received_buf.append(payload)
            if len(received_buf) >= len(data):
                break
            if put_queue != []:
                pub(put_queue.pop())
                await asyncio.sleep(0.001)
            if put_queue != []:
                pub(put_queue.pop())
                await asyncio.sleep(0.001)

    assert data == received_buf


@pytest.mark.xfail(strict=False, reason="flaky depending on platform, middleware ...")
async def test_reliable_extremely_fast(pub: Callable[[bool], None], sub: BaseSub[bool]):
    data = [binary(n) for n in range(30)]
    put_queue: list[bool] = [b for b in data]
    put_queue.reverse()
    received_buf = []
    listener = sub.listen_reliable(fresh=True, queue_size=len(data) * 2)
    pub(put_queue.pop())
    async with afor.soft_timeout(2):
        async for sample in listener:
            payload = int(sample)
            received_buf.append(payload)
            if len(received_buf) >= len(data):
                break
            if put_queue != []:
                pub(put_queue.pop())
            if put_queue != []:
                pub(put_queue.pop())

    assert set(data) == set(received_buf)


async def test_freshness(pub: Callable[[bool], None], sub: BaseSub[bool]):
    payload = True
    new = sub.wait_for_new()
    pub(payload)
    await new
    sample = await afor.soft_wait_for(anext(sub.listen(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample == payload

    new = sub.wait_for_new()
    pub(payload)
    await new
    sample = await afor.soft_wait_for(anext(sub.listen_reliable(fresh=False)), 0.1)
    assert not isinstance(sample, TimeoutError), f"Should get the message"
    assert sample == payload
    await sub.wait_for_value()

    new = sub.wait_for_new()
    pub(payload)
    await new
    sample = await afor.soft_wait_for(anext(sub.listen(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message. got {sample}"

    new = sub.wait_for_new()
    pub(payload)
    await new
    sample = await afor.soft_wait_for(anext(sub.listen_reliable(fresh=True)), 0.1)
    assert isinstance(sample, TimeoutError), f"Should NOT get the message"



import asyncio
import logging
from asyncio.queues import Queue
from contextlib import suppress
from typing import (
    Any,
    AsyncGenerator,
    Awaitable,
    Callable,
    Coroutine,
    Deque,
    Final,
    Generic,
    List,
    Optional,
    Set,
    TypeVar,
)

logger = logging.getLogger(__name__)


class SubClosedException(RuntimeError):
    """Raised when awaiting data from a closed subscriber."""


_MsgType = TypeVar("_MsgType")
_T = TypeVar("_T")


class BaseSub(Generic[_MsgType]):
    def __init__(
        self,
    ) -> None:
        """Abstract (non-specific) asyncio_for_robotics subscriber.

        This defines the different methods to asynchronously get new messages.

        Use ``self.intput_data(your_data)`` to input new messages on this subscriber.

        This object is pure python and not specialize for any transport
        protocol. See asyncio_for_robotics.zenoh.sub for an easy
        implementation.

        To implements your own sheduling methods (new type of queue, buffer,
        genrator ...), please either:

            - inherit from this class, then overide self._input_data_asyncio
            - put a callback inside self.asap_callback
        """
        #: Blocking callbacks called on message arrival
        self.asap_callback: List[Callable[[_MsgType], Any]] = []
        #: Event triggering when the first message is received
        self.alive: asyncio.Event = asyncio.Event()
        #: Number of messages received since start
        self.msg_count: int = 0
        #: Condition that fires on new data
        self.new_value_cond = asyncio.Condition()
        #: queues associated with the generators
        self._dyncamic_queues: Set[Queue[_MsgType]] = set()

        self._event_loop: asyncio.AbstractEventLoop = asyncio.get_event_loop()
        #: Value is available event
        self._value_flag: asyncio.Event = asyncio.Event()
        #: Lastest message value
        self._value: Optional[_MsgType] = None
        #: Event when closing the sub
        self._closed = asyncio.Event()
        self._raise_on_close_exc: Exception = SubClosedException(
            f"Subscriber '{self.name}' has been closed"
        )
        logger.debug("created sub %s", self.name)

    @property
    def name(self) -> str:
        """The friendly name of you subscriber"""
        return "no_name"

    def input_data(self, data: _MsgType) -> bool:
        """Processes incomming data.
        Call this in your subscriber callback.

        .. Note:
            This is threadsafe, thus can run safely on any thread.

        Args:
            data: Data to input on this sub

        Returns:
            False if the even loop has been closed or there is another critical
                problem making this sub unable to work.
        """
        if self._event_loop.is_closed():
            logger.info("Event loop closed, for sub: %s", self.name)
            return False
        self._event_loop.call_soon_threadsafe(self._input_data_asyncio, data)
        return True

    async def wait_for_value(self) -> _MsgType:
        """Latest message.

        Returns:
            The latest message (if none, awaits the first message)
        """
        await self._wait_or_raise_closed(self._value_flag.wait())
        assert self._value is not None
        return self._value

    def wait_for_new(self) -> Coroutine[Any, Any, _MsgType]:
        """Awaits a new value.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the coroutine is awaited

        See wait_for_next to be sure the data is exactly the next received.

        Returns:
            Coroutine holding a message more recent than the one at time of the call
        """
        last_count = self.msg_count

        async def func() -> _MsgType:
            async with self.new_value_cond:
                await self.new_value_cond.wait_for(lambda: self.msg_count > last_count)
            assert self._value is not None
            return self._value

        return self._wait_or_raise_closed(func())

    def wait_for_next(self) -> Coroutine[Any, Any, _MsgType]:
        """Awaits exactly the next value.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the coroutine is awaited

        Returns:
            Coroutine holding the first message received after the call.
        """
        q: asyncio.Queue[_MsgType] = asyncio.LifoQueue(maxsize=2)
        self._dyncamic_queues.add(q)

        async def func() -> _MsgType:
            try:
                val_top = await q.get()
                if q.empty():
                    val_deep = val_top
                else:
                    val_deep = q.get_nowait()
                return val_deep
            finally:
                self._dyncamic_queues.discard(q)

        return self._wait_or_raise_closed(func())

    def listen(self, fresh=False) -> AsyncGenerator[_MsgType, None]:
        """Itterates over the newest message.

        Messages might be skipped, this is not a queue.

        .. Note:
            "Listening" starts the moment this function is called, not when
            the generator is awaited

        Args:
            fresh: If false, first yield can be the latest value

        Returns:
            Async generator itterating over the newest message.
        """
        return self.listen_reliable(fresh, 1, False)

    def listen_reliable(
        self, fresh=False, queue_size: int = 10, lifo=False
    ) -> AsyncGenerator[_MsgType, None]:
        """Itterates over every incomming messages. (does not miss messages)

        .. Note:
            "Listening" starts the moment this function is called, not when
            the generator is awaited.


        Args:
            fresh: If false, first yield can be the latest value
            queue_size: size of the queue of values
            lifo: If True, uses a last in first out queue instead of default fifo.

        Returns:
            Async generator itterating over every incomming message.
        """
        if not lifo:
            q: asyncio.Queue[_MsgType] = asyncio.Queue(maxsize=queue_size)
        else:
            q: asyncio.Queue[_MsgType] = asyncio.LifoQueue(maxsize=queue_size)
        self._dyncamic_queues.add(q)
        logger.debug("Reliable listener primed %s", self.name)
        if self._value_flag.is_set() and not fresh:
            assert self._value is not None, "impossible if flag set"
            q.put_nowait(self._value)
        return self._unprimed_listen_reliable(q)

    async def _unprimed_listen_reliable(
        self, queue: asyncio.Queue
    ) -> AsyncGenerator[_MsgType, None]:
        logger.debug("Reliable listener first iter %s", self.name)
        try:
            while True:
                # logger.debug("Reliable listener waiting data %s", self.name)
                msg = await self._wait_or_raise_closed(queue.get())
                # logger.debug("Reliable listener got data %s", self.name)
                yield msg
                # logger.debug("Reliable listener yielded data %s", self.name)
        finally:
            self._dyncamic_queues.discard(queue)
            logger.debug("Reliable listener closed %s", self.name)

    def _input_data_asyncio(self, msg: _MsgType):
        """Processes incomming data.

        Is only safe to run on the same thread as asyncio.

        Args:
            msg: message
        """
        # logger.debug("Input message %s", self.name)
        for f in self.asap_callback:
            f(msg)
        for q in self._dyncamic_queues:
            if q.full():
                if q.qsize() > 2:
                    logger.warning("Queue full (%s) on %s", q.qsize, self.name)
                q.get_nowait()
            q.put_nowait(msg)
        self._value = msg
        self._value_flag.set()
        self.msg_count += 1
        asyncio.create_task(self._wakeup_new())
        if not self.alive.is_set():
            logger.debug("%s is receiving data", self.name)
            self.alive.set()

    async def _wakeup_new(self):
        """fires new_value_cond"""
        async with self.new_value_cond:
            self.new_value_cond.notify_all()

    async def _wait_or_raise_closed(
        self, awaitable: Coroutine[Any, Any, _T] | Awaitable[_T]
    ) -> _T:
        """Wait for an awaitable or raise exception on subscriber shutdown.

        Raises:
            SubClosedError: if the subscriber is closed before completion.
            Exception: any exception raised by the awaitable.

        Returns:
            Awaited awaitable
        """
        main_task = asyncio.ensure_future(awaitable)
        close_task = asyncio.ensure_future(self._closed.wait())
        done, pending = await asyncio.wait(
            [main_task, close_task],
            return_when=asyncio.FIRST_COMPLETED,
        )
        try:
            if close_task in done:
                logger.info(f"Terminating task because sub is closed")
                raise self._raise_on_close_exc
            return await main_task
        finally:
            for task in pending:
                task.cancel()
            if len(pending) != 0:
                with suppress(asyncio.CancelledError):
                    await asyncio.wait(pending)

    def close(self):
        self._closed.set()


_InType = TypeVar("_InType")
_OutType = TypeVar("_OutType")


class ConverterSub(BaseSub[_OutType]):
    def __init__(
        self,
        sub: BaseSub[_InType],
        convert_func: Callable[[_InType], _OutType] = lambda x: x,
    ) -> None:
        """Subscriber that applies a transformation to another subscriber.

        This subscriber listens reliably to an upstream ``BaseSub`` and
        publishes transformed messages as a new ``BaseSub`` instance. The
        original subscriber is left unchanged.

        Args:
            sub: Upstream subscriber to listen to.
            convert_func: Function applied to each incoming message. Identity
                by default.
        """
        #: Upstream subscriber
        self.sub: BaseSub = sub
        #: Transformation applied to each received message
        self.convert_func = convert_func
        super().__init__()
        #: Background task running the conversion loop
        self._loop_task = asyncio.create_task(self._converter_loop())
        #: Optional callback invoked on close (typically upstream close).
        self.callback_on_close: Optional[Callable[[], Any]] = None
        if hasattr(self.sub, "close"):
            if callable(self.sub.close):
                self.callback_on_close = self.sub.close

    async def _converter_loop(self):
        async for msg in self.sub.listen_reliable():
            new = self.convert_func(msg)
            self._input_data_asyncio(new)

    def close(self):
        self._loop_task.cancel()
        if self.callback_on_close is not None:
            self.callback_on_close()
        super().close()
